# Copyright (c) Stephan Martin <sm@sm-zone.net>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

use strict;

package CERT;

sub new {
   my $that = shift;
   my $class = ref($that) || $that;

   my $self = {};

   bless($self, $class);
}

sub read_certlist {
   my $self    = shift;
   my $main    = shift;
   my $force   = shift;
   my $ca      = $main->{'CA'}->{'actca'};
   my $certdir = $main->{'CA'}->{$ca}->{'dir'}."/certs";
   my $crlfile = $main->{'CA'}->{$ca}->{'dir'}."/crl/crl.pem";

   my($f, $status, $certlist, $crl, $modt);

   $certlist = [];
   
   $modt = (stat($certdir))[9];

   if(defined($self->{'lastread'}) &&
      ($self->{'lastread'} >= $modt) && 
      not defined($force)) {
      return(0);
   }

   if(not defined($main->{'OpenSSL'})) {
      $main->init_openssl($ca);
   }

   $self->{'CRL'} = OpenCA::CRL->new(
      SHELL  => $main->{'OpenSSL'},
      INFILE => $crlfile,
      FORMAT => "PEM"
      );
   
   $crl = $self->{'CRL'}->getParsed();

   opendir(DIR, $certdir) || do {
      $main->print_warning("can't open certdir");
      return(0);
   };

   while($f = readdir(DIR)) {
      next if $f =~ /^\./;

      $f =~ s/\.pem//;
      $status = $self->_get_status(
            $main->{'OpenSSL'}, 
            $f, 
            $main
            );

      my $tmp = MIME::Base64::decode($f);
         $tmp .= "%".$status;

      push(@{$certlist}, $tmp);
   }
   @{$certlist} = sort(@{$certlist});
   closedir(DIR);

   $self->{'certlist'} = $certlist;

   $self->{'lastread'} = time();
   return(1);  # got new list
}

sub revoke_cert {
   my $self = shift;
   my $main = shift;
   my $ca  = $main->{'CA'}->{'actca'};
   my $ind = $main->{'certtab'}->{'list'}->info('selection');

   my($certname, $cert, $status, $certfile, $passwd, $action);

   if(not defined $ind) {
      $main->print_warning("No Certificate Selected!");
      return;
   }

   ($cert, $status) = split(/\%/, $self->{'certlist'}->[$ind]);

   if($status ne 'VALID') {
      $main->print_warning("Can't revoke Certifikate with Status: $status");
      return;
   }

   $certname = MIME::Base64::encode($cert, '');
   $certfile = $main->{'CA'}->{$ca}->{'dir'}."/certs/".$certname.".pem";

   $action = $main->show_cert_revoke_dialog(\$passwd);

   return if $action eq "Cancel";

   if(not defined($main->{'OpenSSL'})) {
      $main->init_openssl($ca);
   }

   if(not $main->{'OpenSSL'}->revoke(
         CAKEY  => $main->{'CA'}->{$ca}->{'dir'}."/cacert.key",
         CACERT => $main->{'CA'}->{$ca}->{'dir'}."/cacert.pem",
         INFILE => $main->{'CA'}->{$ca}->{'dir'}."/certs/".$certname.".pem",
         PASSWD => $passwd
         )) {
      $main->print_warning("revoke failed");
      return;
   }

   $main->{'OpenSSL'}->issueCrl( 
         CAKEY   => $main->{'CA'}->{$ca}->{'dir'}."/cacert.key",
         CACAERT => $main->{'CA'}->{$ca}->{'dir'}."/cacert.pem",
         PASSWD  => $passwd,
         DAYS    => 365,
         OUTFILE => $main->{'CA'}->{$ca}->{'dir'}."/crl/crl.pem",
         OUTFORM => "PEM"
         );

   if (not -s $main->{'CA'}->{$ca}->{'dir'}."/crl/crl.pem") { 
      $main->print_error("issueCrl failed");
   }

   $self->{$certname}->{'status'} = "REVOKED";

   $action = $main->show_key_del_question();

   if($action eq "Yes") {
      unlink($main->{'CA'}->{$ca}->{'dir'}."/keys/".$certname.".pem");
      unlink($main->{'CA'}->{$ca}->{'dir'}."/req/".$certname.".pem");
   }

   # force reread of certlist
   $self->read_certlist($main, 1);

   $main->create_mframe();

   return;
}

sub del_cert {
   my $self = shift;
   my $main = shift;
   my $ca   = $main->{'CA'}->{'actca'};
   my $ind  = $main->{'certtab'}->{'list'}->info('selection');
    
   my($certname, $cert, $certfile, $status);

   if(not defined $ind) {
      $main->print_warning("No Certificate selected!");
      return;
   }

   ($cert, $status) = split(/\%/, $self->{'certlist'}->[$ind]);

   $certname = MIME::Base64::encode($cert, '');
   $certfile = $main->{'CA'}->{$ca}->{'dir'}."/certs/".$certname.".pem";

   if($status eq 'VALID') {
      $main->print_warning("Certificate is not revoked");
      return;
   }

   unlink($certfile);

   # force reread of certlist
   $self->read_certlist($main, 1);

   $main->create_mframe();

   return;
}

sub export_cert {
   my $self = shift;
   my $main = shift;
   my $ca   = $main->{'CA'}->{'actca'};
   my $ind  = $main->{'certtab'}->{'list'}->info('selection');
    
   my($certname, $cert, $certfile, $status, $outfile, $action, 
      $format, $x509, $out, $keyfile, $passwd, $p12passwd);

   if(not defined $ind) {
      $main->print_warning("No Certificate selected");
      return;
   }

   ($cert, $status) = split(/\%/, $self->{'certlist'}->[$ind]);

   $certname = MIME::Base64::encode($cert, '');
   $certfile = $main->{'CA'}->{$ca}->{'dir'}."/certs/".$certname.".pem";

   if(($status eq 'INVALID') || ($status eq 'REVOKED')) {
      $main->print_warning("Certificate seems not to be valid");
      return;
   }

   $action = $main->show_cert_export_dialog(
         \$outfile,
         \$format
         );

   return if $action eq "Cancel";
   return if(not defined($outfile));

   if(not defined($main->{'OpenSSL'})) {
      $main->init_openssl($ca);
   }

   $x509 = OpenCA::X509->new(
         SHELL  => $main->{'OpenSSL'},
         INFILE => $certfile,
         FORMAT => "PEM"
         );

   if($format eq "PEM") {
      $out = $x509->getPEM();
   } elsif ($format eq "DER") {
      $out = $x509->getDER();
   } elsif ($format eq "TXT") {
      $out = $x509->getTXT();
   } elsif ($format eq "P12") {
      $keyfile = $main->{'CA'}->{$ca}->{'dir'}."/keys/".$certname.".pem";
      if(not -s $keyfile) {
         $main->print_warning("Key is necessary for Export as PKCS#12");
         return;
      }

      $action = $main->show_p12_export_dialog(
            \$p12passwd,
            \$passwd
            );

      return if $action eq "Cancel";
      
      unlink($outfile);
      $main->{'OpenSSL'}->dataConvert(
            INFILE    => $certfile,
            KEYFILE   => $keyfile,
            DATATYPE  => 'CERTIFICATE',
            OUTFILE   => $outfile,
            INFORM    => 'PEM',
            OUTFORM   => 'PKCS12',
            PASSWD    => $passwd,
            P12PASSWD => $p12passwd
            );

      if(not -s $outfile) {
         $main->print_warning("dataConvert failed");
         return;
      }

      $main->print_info("Certificate and Key exported to $outfile");
      return;

   } else {
      $main->print_warning("Invalid Format");
      return;
   }

   open(OUT, ">$outfile") || do {
      $main->print_warning("Can't open Outfile");
      return;
   };

   print OUT $out;
   close OUT;
   
   $main->print_info("Certificate exported to: $outfile");

   return;
}

sub parse_cert {
   my $self = shift;
   my $main = shift;
   my $name = shift;
   my($ca, $certfile, $x509, $parsed, $exts, $key, $val, $i, @lines);

   $ca = $main->{'CA'}->{'actca'};

   $certfile = $main->{'CA'}->{$ca}->{'dir'}."/certs/".$name.".pem";

   if(not defined($main->{'OpenSSL'})) {
      $main->init_openssl($ca);
   }

   $x509 = OpenCA::X509->new(
         SHELL  => $main->{'OpenSSL'},
         INFILE => $certfile,
         FORMAT => "PEM"
         );

   $self->{$name}->{'parsed'} = $x509->getParsed();

   # these x.501 attributes are not scanned within OpenCA::REQ
   if($self->{$name}->{'parsed'}->{'DN'} =~ /[\,\/\s]N=([^\,\/\n]+)/i) {
      $self->{$name}->{'parsed'}->{'N'} = $1;
   }
   if($self->{$name}->{'parsed'}->{'DN'} =~ /[\,\/\s]G=([^\,\/\n]+)/i) {
      $self->{$name}->{'parsed'}->{'G'} = $1;
   }
   if($self->{$name}->{'parsed'}->{'DN'} =~ /[\,\/\s]I=([^\,\/\n]+)/i) {
      $self->{$name}->{'parsed'}->{'I'} = $1;
   }
   if($self->{$name}->{'parsed'}->{'DN'} =~ /[\,\/\s]T=([^\,\/\n]+)/i) {
      $self->{$name}->{'parsed'}->{'T'} = $1;
   }
   if($self->{$name}->{'parsed'}->{'DN'} =~ /[\,\/\s]D=([^\,\/\n]+)/i) {
      $self->{$name}->{'parsed'}->{'D'} = $1;
   }

   $self->{$name}->{'EXPDATE'} = _get_date( 
         $self->{$name}->{'parsed'}->{'NOTAFTER'}
         );

   $self->{$name}->{'status'} = $self->_get_status(
         $main->{'OpenSSL'},
         $name,
         $main
         );

   $self->{$name}->{'txt'} = $x509->getTXT();

   @lines = split(/\n/, $self->{$name}->{'txt'});

   # skip everything, we don't want to see
   for($i = 0; $lines[$i] !~ /^[\s\t]*X509v3 extensions:$/i; $i++) {
   }
   $i++;

   while($i < @lines) {
      if($lines[$i] =~ /^[\s\t]*[^:]+:\s*$/) {
         $key = $lines[$i];
         $key =~ s/[\s\t:]*$//g;
         $key =~ s/^[\s\t]*//g;
         $i++;
         $exts->{$key} = [];
         while($lines[$i] !~ /^[\s\t].+:\s*$/ && $i < @lines) {
            $val = $lines[$i];
            $val =~ s/^[\s]+//g;
            $val =~ s/[\s]+$//g;
            $i++;
            next if $val =~ /^$/;
            next if $val =~ /Signature Algorithm:/;
            next if $val =~ /DirName:/;
            push(@{$exts->{$key}}, $val);
         }
      } else {
         $i++;
      }
   }

   $self->{$name}->{'parsed'}->{'exts'} = $exts;

   return;
}

sub _get_status {
   my $self    = shift;
   my $openssl = shift;
   my $name    = shift;
   my $main    = shift;

   my ($ca, $certfile, $x509, $crlfile, $crl, $parsed, $valid, $time);

   $time = time();

   $ca = $main->{'CA'}->{'actca'};
   
   $self->{'CRL'} = OpenCA::CRL->new( 
         SHELL  => $openssl,
         INFILE => $main->{'CA'}->{$ca}->{'dir'}."/crl/crl.pem",
         FORMAT => "PEM"
         );
   
   $crl = $self->{'CRL'}->getParsed();

   if(not defined($self->{$name}->{'parsed'})) {
      $self->parse_cert($main, $name);
   }

   $parsed = $self->{$name}->{'parsed'};

   $crl->{'ISSUER'} =~ s/^\///;
   $crl->{'ISSUER'} =~ s/, /\//g;
   $parsed->{'ISSUER'} =~ s/^\///;
   $parsed->{'ISSUER'} =~ s/, /\//g;

   if($self->{$name}->{'EXPDATE'} < $time) {
      $self->{$name}->{'status'} = "EXPIRED";
      # keep database up to date
      if($crl->{'ISSUER'} eq $parsed->{'ISSUER'}) {
         _set_expired($parsed->{'SERIAL'}, $main);
      }
      return("EXPIRED");
   }
   
   if(($crl->{'ISSUER'} ne $parsed->{'ISSUER'}) || 
      (not defined($parsed->{'ISSUER'}))) {
      print "DEBUG: Issuers differ:\n"
           ."   crl-issuer: $crl->{'ISSUER'}\n"
           ."   crt-issuer: $parsed->{'ISSUER'}\n";
   
      $self->{$name}->{'status'} = "UNKNOWN";
      return("UNKNOWN");
   }
   
   foreach my $revoked (@{$crl->{'LIST'}}) {
      next if ($parsed->{'SERIAL'} ne $revoked->{'SERIAL'});
      if ($parsed->{'SERIAL'} == $revoked->{'SERIAL'}) {
         $self->{$name}->{'status'} = "REVOKED";
         return("REVOKED");
      }
   }

   $self->{$name}->{'status'} = "VALID";
   return("VALID");
}

sub _set_expired {
   my $serial = shift;
   my $main   = shift;
   
   my $ca = $main->{'CA'}->{'actca'};

   my $index = $main->{'CA'}->{$ca}->{'dir'}."/index.txt";

   open(IN, "<$index") || do {
      $main->print_warning("can't open index");
      return;
   };

   my @lines = <IN>;

   close IN;

   open(OUT, ">$index") || do {
      $main->print_warning("can't open index");
      return;
   };

   foreach my $l (@lines) {
      if($l =~ /\t$serial\t/) {
         $l =~ s/^V/E/;
      }
      print OUT $l;
   }

   close OUT;

   return;
}

sub _get_date {
   my $string = shift;

   $string =~ s/  / /g;

   my @t1 = split(/ /, $string);
   my @t2 = split(/:/, $t1[2]);

      $t1[0] = _get_index($t1[0]);

   my $ret = Time::Local::timelocal($t2[2],$t2[1],$t2[0],$t1[1],$t1[0],$t1[3]);

   return($ret);
}

sub _get_index {
   my $m = shift;
   my @a = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);

   for(my $i = 0; $a[$i]; $i++) {
      return $i if($a[$i] eq $m);
   }
}

1
